import SkribblClient from "../logic/SkribblClient.js";
import {CustomElement} from "../util/Util.js";
import SkribblCanvas from "./SkribblCanvas.js";

/**
 * Custom element `<skribbl-canvas-container>` that manages the canvas and its overlays.
 */
export default class SkribblCanvasContainer extends CustomElement {
	/**
	 * @param {SkribblClient} client 
	 */
	constructor(client){
		super();
		this.attachShadow({mode:"open"});
		this.shadowRoot.innerHTML = `
			<style>
				:host {
					display: block;
					position: relative;
				}
				svg {
					display: block;
					width: 100%;
				}
				#container {
					position: absolute;
					left: 0;
					top: 0;
					width: 100%;
					height: 100%;
					overflow: hidden;
					background: #ffffff;
					border-radius: 5px;
				}
				#controls {
					position: absolute;
					top: 100%;
					margin-top: 10px;
				}
				#controls.hidden {
					display: none;
				}
			</style>
			<!-- invisible svg to maintain an aspect ratio of 960:720, aka 4:3 -->
			<svg viewBox="0 0 960 720"></svg>
			<div id="container">
				<slot name="overlay"></slot>
			</div>
			<div id="controls">
				Color settings & stuff goes here.
			</div>
		`;
		this._canvas = new SkribblCanvas(client);
		this._container = this.shadowRoot.getElementById("container");
		this._container.insertBefore(this._canvas,this._container.firstChild);
		this._controls = this.shadowRoot.getElementById("controls");
		client.state.onChange(state=>{
			/** whether the player should currently be allowed to draw on the canvas */
			let allowToDraw = state.hasGameStarted&&state.drawingPlayer==state.playerIndex&&typeof state.word=="string";
			this._canvas.drawable = allowToDraw;
			this._controls.classList.toggle("hidden",!allowToDraw);
		},{onlyWhen:this.connected});
		// TODO implement color settings and set the correspondig properties of `this._canvas` whenever they change here.
	}

	set overlay(overlay){
		if (this._overlay){
			this._overlay.remove();
		}
		this._overlay = overlay;
		if (overlay){
			overlay.slot = "overlay";
			this.appendChild(overlay);
		}
	}

	/**
	 * @type {SkribblCanvasOverlay}
	 */
	get overlay(){
		return this._overlay;
	}
}
/**
 * Custom element `<skribbl-canvas-overlay>` for overlays on the `<skribbl-canvas>`-element.
 */
export class SkribblCanvasOverlay extends HTMLElement {
	/**
	 * Constructs a new SkribblCanvasOverlay and appends the given nodes or strings to it.
	 * @param {...(string|Node)} nodes
	 */
	constructor(...nodes){
		super();
		this.attachShadow({mode:"open"});
		this.shadowRoot.innerHTML = `
			<style>
				:host {
					display: block;
					position: absolute;
					top: 0;
					right: 0;
					bottom: 0;
					left: 0;
					background: #000000bf;
					color: #ffffff;
				}
				#contents {
					position: absolute;
					left: 50%;
					top: 50%;
					max-width: 80%;
					transform: translate(-50%,-50%);
					text-align: center;
					font-size: 1.3em;
				}
				::slotted(button) {
					padding: 3px 5px;
					margin: 3px;
					border: none;
					border-radius: 3px;
					background: #dfdfdf;
					color: #000000;
					font-size: 1em;
				}
				::slotted(button:hover) {
					cursor: pointer;
				}
			</style>
			<div id="contents">
				<slot></slot>
			</div>
		`;
		this.append(...nodes);
	}
}
customElements.define("skribbl-canvas-container",SkribblCanvasContainer);
customElements.define("skribbl-canvas-overlay",SkribblCanvasOverlay);