/** @type {WeakMap<Element,()=>void>} */
const callbackMap = new WeakMap();
const observer = new IntersectionObserver((entries)=>{
	entries.forEach(entry=>{
		if (entry.isIntersecting){
			callbackMap.get(entry.target)();
			callbackMap.delete(entry.target);
			observer.unobserve(entry.target);
		}
	});
});
/**
 * Registers a callback to be called when the given element enters the viewport and becomes visible for the first time.
 * @param {Element} element 
 * @param {()=>void} callback 
 */
export function onFirstVisible(element,callback){
	callbackMap.set(element,callback);
	observer.observe(element);
}
/**
 * @typedef {T extends any?{readonly [K in keyof T]:DeepReadOnly<T[K]>}:T} DeepReadOnly
 * @template {unknown} T
 */
/**
 * Freezes the object and all of its values using `Object.freeze(obj)`, effectively making the entire object immutable.
 * @template {unknown} T
 * @param {T} obj
 * @return {DeepReadOnly<T>}
 */
export function deepFreeze(obj){
	/** @type {unknown[]} */
	let objects = [];
	/** @param {unknown} obj */
	let freeze = obj=>{
		if (!objects.includes(obj)){
			objects.push(obj);
			Object.freeze(obj);
			Object.values(obj).forEach(subObj=>freeze(subObj));
		}
	};
	// @ts-ignore
	return obj;
}
/**
 * A collection of callback functions.
 * @template {any[]} Args argument types of the callbacks
 */
export class CallbackHandler {
	constructor(){
		/** @type {((...args:Args)=>void)[]} */
		this._callbacks = [];
	}

	/**
	 * Adds a callback to be called whenever `callAll(...args)` is called.
	 * @param {(...args:Args)=>void} callback
	 * @param {object} [options]
	 * @param {Value<boolean>|ReadOnlyValue<boolean>} [options.onlyWhen] when specified, calls the callback only when this is value is true.
	 */
	addCallback(callback,{onlyWhen=null}={}){
		if (onlyWhen==null){
			this._callbacks.push(callback);
		}else{
			onlyWhen.onChange(isTrue=>{
				if (isTrue){
					this._callbacks.push(callback);
				}else{
					this._callbacks = this._callbacks.filter(c=>c!=callback);
				}
			});
		}
	}

	/**
	 * Removes a callback registered using `addCallback(callback)`. Doesn't work if the callback was registered with the `onlyWhen` option specified.
	 * @param {(...args:Args)=>void} callback
	 */
	removeCallback(callback){
		this._callbacks = this._callbacks.filter(c=>c!=callback);
	}

	/**
	 * @param {Args} args
	 */
	callAll(...args){
		this._callbacks.slice().forEach(callback=>callback(...args));
	}
}
/**
 * A value of any type together with a `onChange(callback)` method.
 * @template T
 */
export class Value {
	/**
	 * @param {T} value 
	 */
	constructor(value){
		this._value = value;
		/** @type {CallbackHandler<[value:T]>} */
		this._onChangeCallbacks = new CallbackHandler();
	}
	
	set value(value){
		if (this._value!==value){
			this._value = value;
			this._onChangeCallbacks.callAll(value);
		}
	}

	get value(){
		return this._value;
	}

	/**
	 * A readonly version of this object.
	 * @readonly
	 */
	get readOnly(){
		if (!this._readOnly){
			/** @type {ReadOnlyValue<T>} */
			this._readOnly = new ReadOnlyValue(this);
		}
		return this._readOnly;
	}
	
	/**
	 * Registers a callback to be called whenever the value changes, and once immediatly, unless `onRegister` is explicitly set to false.
	 * @param {(value:T)=>void} callback
	 * @param {object} [options]
	 * @param {Value<boolean>|ReadOnlyValue<boolean>} [options.onlyWhen] when specified, calls the callback only when this is value is true.
	 * @param {boolean} [options.onRegister] whether or not to call the callbacks once immediately. defauls to `true`
	 */
	onChange(callback,{onlyWhen=null,onRegister=true}={}){
		this._onChangeCallbacks.addCallback(callback,{onlyWhen});
		if (onlyWhen!=null){
			let prevValue = this._value;
			onlyWhen.onChange(isTrue=>{
				if (isTrue){
					if (prevValue!==this._value){
						callback(this._value);
					}
				}else{
					prevValue = this._value;
				}
			});
		}
		if (onRegister){
			callback(this._value);
		}
	}
}
/**
 * A wrapper for a `Value` without a `.value` setter method, whose value can thus only be changed via the original `Value` object that was passed to the constructor.
 * @template {*} T
 */
export class ReadOnlyValue {
	/**
	 * @param {Value<T>} value
	 */
	constructor(value){
		this._value = value;
	}

	get value(){
		return this._value.value;
	}
	
	/**
	 * Registers a callback to be called whenever the value changes, and once immediatly, unless `onRegister` is explicitly set to false.
	 * @param {(value:T)=>void} callback
	 * @param {object} [options]
	 * @param {Value<boolean>|ReadOnlyValue<boolean>} [options.onlyWhen] when specified, calls the callback only when this is value is true.
	 * @param {boolean} [options.onRegister] whether or not to call the callbacks once immediately. defauls to `true`
	 */
	onChange(callback,{onlyWhen=null,onRegister=true}={}){
		this._value.onChange(callback,{onlyWhen,onRegister});
	}
}
/**
 * Base class for custom elements with a few utility methods.
 */
export class CustomElement extends HTMLElement {
	constructor(){
		super();
		this._connected = new Value(false);
	}

	/**
	 * Whether the element is currently connected to the document, like `Node.isConnected`, but wrapped in a `Value` object, so it can be monitored using the `onChange(callback)` method.
	 */
	get connected(){
		return this._connected.readOnly;
	}

	connectedCallback(){
		this._connected.value = true;
	}

	disconnectedCallback(){
		this._connected.value = false;
	}
}